from ball_control_sim import BallControlSim
from ch4_state_manager import Ch4StateManager
from scenario_control import ScenarioControl
from update_reporter import UpdateReporter

class BallControlCh4(BallControlSim, ScenarioControl):

    def __init__(self, update_reporter: UpdateReporter, delay_multiplier: float = 1.0):
        super().__init__(update_reporter=update_reporter, delay_multiplier=delay_multiplier)
        self.state_manager = Ch4StateManager()



import asyncio
from dataclasses import replace

from ball_control import BallControl
from scenario import Scenario, ScenarioProgress
from state_manager import StateManager
from scenario_control import ScenarioControl
from state_update_model import StateModel, StatePosition, StateUpdateModel, get_default_state
from update_reporter import UpdateReporter

class BallControlSim(BallControl, ScenarioControl):

    delay_mult: float
    update_reporter: UpdateReporter
    state_manager: StateManager
    state: StateModel

    def __init__(self, update_reporter: UpdateReporter, delay_multiplier: float = 1.0):
        self.update_reporter = update_reporter
        self.state_manager = StateManager()
        self.delay_mult = delay_multiplier
        self.state = get_default_state()

    async def __aenter__(self):
        pass
    
    async def __aexit__(self, *_):
        await self.update_reporter.shutdown()

    async def __send_update(self, include_balls: bool = False, include_dimensions: bool = False):
        state_to_send = self.state if include_balls else replace(self.state, balls = None)
        state_to_send = state_to_send if include_dimensions else replace(state_to_send, max_x = 0, max_y = 0)

        state_update: StateUpdateModel = StateUpdateModel(
                userId="glen",
                state=state_to_send,
                delay_multiplier=self.delay_mult
            )

        await self.update_reporter.send_update(state_update)
   
    async def _delay(self, duration: float):
        await asyncio.sleep(duration * self.delay_mult)

    async def _move_relative(self, x: int, y: int, delay: float = 1.0):
        delayTask = asyncio.create_task(self._delay(delay))
        await self.__send_update()
        await delayTask

    async def move_horizontally(self, distance: int):
        if (0 == distance):
            return
        
        try:
            self.state = self.state_manager.move_horizontally_start(state=self.state, distance=distance)
            await self._move_relative(x=distance, y=0, delay=1.0)            
        finally:
            self.state = self.state_manager.move_horizontally_end(state=self.state)
            await self.__send_update()

    async def move_vertically(self, distance: int) -> None:
        if (0 == distance):
            return
        
        try:
            self.state = self.state_manager.move_vertically_start(state=self.state, distance=distance)
            await self._move_relative(x=0, y=distance, delay=1.5)
        finally:
            self.state = self.state_manager.move_vertically_end(state=self.state)
            await self.__send_update()

    def get_position(self) -> StatePosition:
        return self.state.claw.pos

    async def open_claw(self):
        try:
            delayTask = asyncio.create_task(self._delay(0.3))
            self.state = self.state_manager.open_claw_start(state=self.state)

            await self.__send_update()
            await delayTask
        finally:
            self.state = self.state_manager.open_claw_end(state=self.state)

    async def close_claw(self):
        try:
            delayTask = asyncio.create_task(self._delay(0.3))
            self.state = self.state_manager.close_claw_start(state=self.state)

            await self.__send_update()
            await delayTask
        finally:
            self.state = self.state_manager.close_claw_end(state=self.state)

    async def set_scenario(self, scenario: Scenario):
        self.state = self.state_manager.set_scenario(state=self.state, scenario=scenario)
        await self.__send_update(include_balls = True, include_dimensions = True)

    def get_progress(self) -> ScenarioProgress:
        if not self.state_manager.scenario:
            return ScenarioProgress(completed=0, total=0)
        return self.state_manager.scenario.get_progress(self.state)

    def is_in_goal_state(self) -> bool:
        if not self.state_manager.scenario:
            return False
        return self.state_manager.scenario.is_in_goal_state(self.state)
    
    def get_state(self) -> StateModel:
        return self.state


from dataclasses import dataclass, replace
from scenario import Scenario
from state_update_model import (
    StateBall,
    StateModel,
    StatePosition,
    get_default_state,
)


@dataclass
class Ch0Scenario(Scenario):
    """Challenge Implementation"""

    def get_goal_state_description(self) -> str:
        return f"All marbles shall be in the leftmost column.\n{self.get_dimensions_description()}"
    
    def get_initial_state(self) -> StateModel:
        balls = [
            StateBall(pos=StatePosition(x=3, y=4), color="blue"),
            StateBall(pos=StatePosition(x=2, y=4), color="blue"),
            StateBall(pos=StatePosition(x=1, y=4), color="green"),
        ]
        return replace(get_default_state(), balls = balls)

    def is_in_goal_state(self, state: StateModel) -> bool:
        
        columns: list[list[StateBall]] = [[] for _ in range(state.max_x + 1)]
        for ball in state.balls:
            columns[ball.pos.x].append(ball)

        # No ball in claw
        if state.claw.ball_color:
            return False

        # 3 balls in leftmost column
        if len(columns[0]) != 3:
            return False

        return True
    


from dataclasses import dataclass

MIN_X = 0
MIN_Y = 0

@dataclass
class StatePosition:
    x: int
    y: int

    def __str__(self) -> str:
        return f"x={self.x} y={self.y}"

@dataclass
class StateBall:
    pos: StatePosition
    color: str
    value: int = 0
    label: str = ""

    def __str__(self) -> str:
        return f"{self.color} {self.label}"

@dataclass
class Claw:
    pos: StatePosition
    open: bool
    ball_color: str
    ball_value: int
    ball_label: str

@dataclass
class Spotlight:
    on: bool
    pos: StatePosition

@dataclass
class Highlight:
    xMin: int
    xMax: int
    yMin: int
    yMax: int
    color: str

@dataclass
class StateModel:
    max_x: int
    max_y: int
    balls: list[StateBall]
    claw: Claw
    isInGoalState: bool
    moving_horizontally: bool
    moving_vertically: bool
    operating_claw: bool
    spotlight: Spotlight | None
    highlights: list[Highlight] | None

@dataclass
class StateUpdateModel:
    userId: str
    state: StateModel
    delay_multiplier: float


def get_default_state() -> StateModel:
    return StateModel(
        max_x=3,
        max_y=4,
        balls=[StateBall(pos=StatePosition(x=2, y=4), color="blue")],
        claw=Claw(pos=StatePosition(x=0, y=0), open=True, ball_color="", ball_value=0, ball_label=""),
        isInGoalState=False,
        moving_horizontally=False,
        moving_vertically=False,
        operating_claw=False,
        spotlight=None,
        highlights=None
    )



import asyncio
import sys

sys.path.append("../content/ballsortutils")

from ball_control import BallControl
from state_update_model import StatePosition

async def move_ball(bc: BallControl, src: StatePosition, dest: StatePosition):    
    rel_x = src.x - bc.get_position().x
    rel_y = src.y - bc.get_position().y
    await asyncio.gather(
        bc.move_horizontally(rel_x),
        bc.move_vertically(rel_y),
        bc.open_claw())
    await bc.close_claw()
    
    rel_x = dest.x - bc.get_position().x
    rel_y = dest.y - bc.get_position().y
    await asyncio.gather(
        bc.move_horizontally(rel_x),
        bc.move_vertically(rel_y))
    await bc.open_claw()

async def move_ball_by_column(bc: BallControl, src_x: int, dest_x: int):
    src_column_top_occupied_y = min([ball.pos.y for ball in bc.get_state().balls if ball.pos.x == src_x],default=bc.get_state().max_y)
    dest_column_top_vacant_y = min([ball.pos.y for ball in bc.get_state().balls if ball.pos.x == dest_x],default=bc.get_state().max_y + 1) - 1
    await move_ball(bc=bc, src=StatePosition(x=src_x, y=src_column_top_occupied_y), dest=StatePosition(x=dest_x, y=dest_column_top_vacant_y))


#from ball_control_ch2 import BallControlCh2
import ball_control_ch2
from ball_control_ch4 import BallControlCh4
from ball_control_sim import BallControlSim
#from v1_update_reporter import V1UpdateReporter
#from ably_update_reporter import AblyUpdateReporter
#from ably_rest_update_reporter import AblyRestUpdateReporter
from postmessage_update_reporter import PostMessageUpdateReporter
#from dummy_ur import DummyUpdateReporter

def get_control_sim(delay_multiplier: float = 1.0) -> BallControlSim:
    #reporter = V1UpdateReporter()
    #reporter = AblyUpdateReporter()
    #reporter = AblyRestUpdateReporter()
    reporter = PostMessageUpdateReporter()
    #reporter = DummyUpdateReporter()
    
    return BallControlSim(update_reporter=reporter, delay_multiplier=delay_multiplier)

def get_ch2_control_sim(delay_multiplier: float = 1.0) -> ball_control_ch2.BallControlCh2:
    reporter = PostMessageUpdateReporter()
    return ball_control_ch2.BallControlCh2(update_reporter=reporter, delay_multiplier=delay_multiplier)

def get_ch4_control_sim(delay_multiplier: float = 1.0) -> BallControlCh4:
    reporter = PostMessageUpdateReporter()
    return BallControlCh4(update_reporter=reporter, delay_multiplier=delay_multiplier)

def get_ch5_control_sim(delay_multiplier: float = 1.0) -> BallControlCh4:
    return get_ch4_control_sim(delay_multiplier=delay_multiplier)


from scenario import ScenarioProgress
from state_update_model import StateModel, StatePosition


class IllegalBallControlStateError(Exception):
    "Raised when a control command is issued while a command for movement along the same axis is still being executed"
    pass

class BallControl(object):
    """Interface for controlling a ball manipulator in a grid"""

    async def __aenter__(self):
        pass
    
    async def __aexit__(self, *_):
        pass

    async def move_horizontally(self, distance: int):
        """Move the claw horizontally"""
        pass

    async def move_vertically(self, distance: int):
        """Move the claw vertically"""
        pass

    async def open_claw(self):
        pass

    async def close_claw(self):
        pass

    def get_position(self) -> StatePosition:
        raise NotImplementedError
    
    def get_progress(self) -> ScenarioProgress:
        """Returns progress toward goal state."""
        raise NotImplementedError

    def is_in_goal_state(self) -> bool:
        """Returns true only if state fulfills the goal state criteria."""
        raise NotImplementedError
    
    def get_state(self) -> StateModel:
        """Returns the complete state of the board."""
        raise NotImplementedError



from dataclasses import dataclass, replace
from scenario import Scenario
from state_update_model import (
    StateBall,
    StateModel,
    StatePosition,
    get_default_state,
)


@dataclass
class Ch1Scenario(Scenario):
    """Challenge Implementation"""

    def get_goal_state_description(self) -> str:
        return f"Yellow and pink marbles in separate columns.\n{self.get_dimensions_description()}"
    
    def get_initial_state(self) -> StateModel:
        balls = [
            StateBall(pos=StatePosition(x=0, y=4), color="pink"),
            StateBall(pos=StatePosition(x=0, y=3), color="yellow"),
            StateBall(pos=StatePosition(x=0, y=2), color="pink"),
            StateBall(pos=StatePosition(x=1, y=4), color="yellow"),
        ]
        return replace(get_default_state(), balls = balls)

    def is_in_goal_state(self, state: StateModel) -> bool:
        
        columns: list[list[StateBall]] = [[] for _ in range(state.max_x+1)]
        for ball in state.balls:
            columns[ball.pos.x].append(ball)

        # No ball in claw
        if state.claw.ball_color:
            return False

        # number of non-empty columns shall be 2
        if len([column for column in columns if len(column) != 0]) != 2:
            return False
        
        # number of columns with mixed color balls shall be 0
        if len([column for column in columns if len(set([ball.color for ball in column])) > 1]) != 0:
            return False
        
        return True
    
    


from state_update_model import (
    StateBall,
    StateModel,
    StatePosition,
)


def is_ball_in_claw(state: StateModel) -> bool:
    return not not state.claw.ball_color


def is_ball_at_current_pos(state: StateModel) -> bool:
    return next(
        (True for ball in state.balls if ball.pos == state.claw.pos),
        False,
    )
    # balls_at_current_pos = filter(lambda ball: ball.pos == self.state.claw.pos, self.state.balls)
    # return any(balls_at_current_pos)

def get_ball_at(state: StateModel, pos: StatePosition) -> StateBall | None:
    return next(
        (ball for ball in state.balls if pos == ball.pos),
        None,
    )

def get_ball_at_current_pos(state: StateModel) -> StateBall | None:
    return next(
        (ball for ball in state.balls if ball.pos == state.claw.pos),
        None,
    )

def get_top_occupied_index(state: StateModel) -> int:
    y_indexes_in_current_column = [
        ball.pos.y for ball in state.balls if ball.pos.x == state.claw.pos.x
    ]
    top_occupied_y_index = (
        min(y_indexes_in_current_column) if y_indexes_in_current_column else state.max_y + 1
    )
    return top_occupied_y_index


def get_top_vacant_index(state: StateModel) -> int:
    return get_top_occupied_index(state) - 1



from ball_control import IllegalBallControlStateError
from ball_control_sim import BallControlSim
from ch2_state_manager import Ch2StateManager
from scenario_control import ScenarioControl
from state_update_model import StatePosition
from update_reporter import UpdateReporter

class BallControlCh2(BallControlSim, ScenarioControl):

    ch2_state_manager: Ch2StateManager

    def __init__(self, update_reporter: UpdateReporter, delay_multiplier: float = 1.0):
        super().__init__(update_reporter=update_reporter, delay_multiplier=delay_multiplier)
        self.ch2_state_manager = Ch2StateManager()

    async def read_scales(self) -> int:
        """
        Returns negative if left is heavier, positive if right is heavier, 0 if equal.
        Left scale is the sum of ball weights in columns 2
        Right scale is the sum of ball weights in columns 3
        """

        left_pos = StatePosition(x=2, y=4)
        right_pos = StatePosition(x=3, y=4)
        
        if (self.state.claw.pos == left_pos or self.state.claw.pos == right_pos) and self.state.operating_claw:
            raise IllegalBallControlStateError("Scales can not be used while claw is opening or closing.")
        
        await self._delay(0.3)

        # Hard coded weights
        def get_weight_by_color(color: str) -> int:
            match color:
                case "yellow":
                    return 5
                case "green":
                    return 2
                case "blue":
                    return 1
                case _:
                    return 0
                
        def get_column_weight(x: int) -> int:
            return sum([get_weight_by_color(ball.color) for ball in self.state.balls if ball.pos.x == x])
        
        return get_column_weight(3) - get_column_weight(2)



from dataclasses import dataclass
from ball_control import IllegalBallControlStateError
from state_utils import (
    get_top_occupied_index,
    get_top_vacant_index,
    is_ball_at_current_pos,
    is_ball_in_claw,
)
from state_update_model import (
    StateModel,
    MIN_X,
    MIN_Y,
)


@dataclass
class StateValidator:
    """Validates operations"""

    def move_horizontally(self, state: StateModel, distance: int):
        if (state.moving_horizontally):
            raise IllegalBallControlStateError("Already moving horizontally")
        
        newX = state.claw.pos.x + distance
        if newX < MIN_X or newX > state.max_x:
            raise IllegalBallControlStateError(f"X coordinate out of bounds x={newX} minX={MIN_X} maxX={state.max_x}")
    
    def move_vertically(self, state: StateModel, distance: int) -> None:
        if (state.moving_vertically):
            raise IllegalBallControlStateError("Already moving vertically")
        
        newY = state.claw.pos.y + distance
        if newY < MIN_Y or newY > state.max_y:
            raise IllegalBallControlStateError(f"Y coordinate out of bounds y={newY} minY={MIN_Y} maxY={state.max_y}")

    def _check_claw_for_ongoing_operations(self, state: StateModel):
        if (state.operating_claw):
            raise IllegalBallControlStateError("Claw already opening or closing")
        
        if (state.moving_horizontally or state.moving_vertically):
            raise IllegalBallControlStateError("Marble dropped while claw is in motion")

    def open_claw(self, state: StateModel):
        #print("state: ", state)
        if not is_ball_in_claw(state):
            return
        
        self._check_claw_for_ongoing_operations(state=state)

        if state.claw.pos.y != get_top_vacant_index(state):
            raise IllegalBallControlStateError(
                f"Illegal drop location. Must be topmost vacant position ({get_top_vacant_index(state)}). Y={state.claw.pos.y}."
            )

    def close_claw(self, state: StateModel):
        #print("state: ", state)
        if not is_ball_at_current_pos(state):
            return
        
        self._check_claw_for_ongoing_operations(state=state)

        if state.claw.pos.y != get_top_occupied_index(state):
            raise IllegalBallControlStateError(
                f"Illegal grab. Must be topmost marble position ({get_top_occupied_index(state)}). Y={state.claw.pos.y}."
            )



from dataclasses import dataclass
from ball_control import IllegalBallControlStateError
from state_utils import (
    is_ball_in_claw,
)
from state_update_model import (
    StateModel,
)
from state_validator import StateValidator


@dataclass
class Ch4StateValidator(StateValidator):
    """Validates operations"""

    def open_claw(self, state: StateModel):
        super().open_claw(state=state)

        if not is_ball_in_claw(state):
            return

        if state.claw.pos.y == state.max_y:
            return

        ball_below = next(
            (
                ball
                for ball in state.balls
                if ball.pos.x == state.claw.pos.x and ball.pos.y == state.claw.pos.y + 1
            ),
            None,
        )
        if not ball_below:
            return

        # check for values of balls
        if ball_below.value < state.claw.ball_value:
            raise IllegalBallControlStateError(
                f"Marble (value={state.claw.ball_value})dropped on top of lower value ({ball_below.value}) marble"
            )



from state_update_model import StateUpdateModel

class UpdateReporter(object):
    """Interface for reporting state"""

    async def send_update(self, stateUpdate: StateUpdateModel):
        """Report state update"""
        pass

    async def shutdown(self):
        """Any cleanup to do before the object is disposed"""
        pass


from IPython.display import display, Javascript

def open_bs_window(url="https://aheed.github.io/ballsort/"):
    dist_obj = Javascript(f"""
        const bsWindow = window.open('{url}', 'bswindow', 'height=700, width=600');
        window.bswin = bsWindow.window;
        """);
    display(dist_obj)



from dataclasses import dataclass, replace
from scenario import Scenario
from state_update_model import (
    StateBall,
    StateModel,
    StatePosition,
    get_default_state,
)


@dataclass
class Ch3Scenario(Scenario):
    """Challenge Implementation"""

    def get_goal_state_description(self) -> str:
        return f"Turn Polish flag 🇵🇱 into Indonesian flag 🇮🇩.\n{self.get_dimensions_description()}"
    
    def get_initial_state(self) -> StateModel:
        
        def build_column(x: int) -> list[StateBall]:
            r_balls = [StateBall(pos=StatePosition(x=x, y=y), color="red") for y in range(3, 5)]
            w_balls = [StateBall(pos=StatePosition(x=x, y=y), color="white") for y in range(1, 3)]
            return r_balls + w_balls
        
        balls: list[StateBall] = []
        max_x = 4
        for x in range(max_x): #intentionally leave rightmost column empty
            balls = balls + build_column(x=x)
        
        return replace(get_default_state(), balls = balls, max_x=max_x)

    def is_in_goal_state(self, state: StateModel) -> bool:

        # No ball in claw
        if state.claw.ball_color:
            return False
        
        def validate_ball(ball: StateBall) -> bool:
            # no balls in rightmost column
            if ball.pos.x > 3:
                return False

            # no higher column than 4    
            if ball.pos.y < 1:
                return False

            # all red balls on high y coordinates    
            if ball.color == "red" and ball.pos.y > 2:
                return False

            # all white balls on low y coordinates    
            if ball.color == "white" and ball.pos.y < 3:
                return False

            return True

        return not next((True for ball in state.balls if not validate_ball(ball)), False)
    
    


import asyncio
import json
from dataclasses import asdict
from update_reporter import UpdateReporter
from state_update_model import StateUpdateModel
from IPython.display import display,Javascript



class PostMessageUpdateReporter(UpdateReporter):
    """UpdateReporter using window.postMessage to push updates"""

    client_lock = asyncio.Lock()

    async def send_update(self, stateUpdate: StateUpdateModel):
        async with self.client_lock:
            stringified_obj = json.dumps(asdict(stateUpdate))
            display_obj = Javascript(f"""
                var existingWin = window.bswin;
                existingWin && existingWin.postMessage('{stringified_obj}', "*");    
                """);
            display(display_obj)

    async def shutdown(self):
        pass
